import { type IncomingMessage, type ServerResponse } from 'node:http';
import { AsyncLocalStorage } from 'node:async_hooks';
import { EggCore, Request as EggCoreRequest, Response as EggCoreResponse, Router } from '@eggjs/core';
import type { EggCoreOptions, Next, MiddlewareFunc as EggCoreMiddlewareFunc, ILifecycleBoot } from '@eggjs/core';
import { EggContextLogger as ContextLogger, EggLoggers, EggLogger } from 'egg-logger';
import { Cookies as ContextCookies } from '@eggjs/cookies';
import type { Agent } from './agent.js';
import type { Application } from './application.js';
import Context from '../app/extend/context.js';
import type { EggAppConfig } from './types.js';
import { IMessenger } from './core/messenger/index.js';
import { ContextHttpClient } from './core/context_httpclient.js';
import { HttpClient, type HttpClientRequestOptions, type HttpClientRequestURL, type HttpClientResponse, type HttpClientOptions } from './core/httpclient.js';
import { BaseContextClass } from './core/base_context_class.js';
import { BaseHookClass } from './core/base_hook_class.js';
import type { EggApplicationLoader } from './loader/index.js';
declare const EGG_PATH: unique symbol;
export interface EggApplicationCoreOptions extends Omit<EggCoreOptions, 'baseDir'> {
    mode?: 'cluster' | 'single';
    clusterPort?: number;
    baseDir?: string;
}
export declare class Request extends EggCoreRequest {
    app: EggCore;
    response: Response;
    ctx: Context;
}
export declare class Response extends EggCoreResponse {
    app: EggCore;
    request: Request;
    ctx: Context;
}
export type { ILifecycleBoot, ILifecycleBoot as IBoot, Next, };
export type EggContext = Context;
export type MiddlewareFunc<T extends Context = Context> = EggCoreMiddlewareFunc<T>;
export { Context, Router, };
/**
 * Based on koa's Application
 * @see https://github.com/eggjs/egg-core
 * @see https://github.com/eggjs/koa/blob/master/src/application.ts
 * @augments EggCore
 */
export declare class EggApplicationCore extends EggCore {
    #private;
    ctxStorage: AsyncLocalStorage<Context>;
    ContextCookies: typeof ContextCookies;
    ContextLogger: typeof ContextLogger;
    ContextHttpClient: typeof ContextHttpClient;
    HttpClient: typeof HttpClient;
    HttpClientNext: typeof HttpClient;
    /**
     * Retrieve base context class
     * @member {BaseContextClass} BaseContextClass
     * @since 1.0.0
     */
    BaseContextClass: typeof BaseContextClass;
    /**
     * Retrieve base controller
     * @member {Controller} Controller
     * @since 1.0.0
     */
    Controller: typeof BaseContextClass;
    /**
     * Retrieve base service
     * @member {Service} Service
     * @since 1.0.0
     */
    Service: typeof BaseContextClass;
    /**
     * Retrieve base subscription
     * @member {Subscription} Subscription
     * @since 2.12.0
     */
    Subscription: typeof BaseContextClass;
    /**
     * Retrieve base context class
     * @member {BaseHookClass} BaseHookClass
     */
    BaseHookClass: typeof BaseHookClass;
    /**
     * Retrieve base boot
     * @member {Boot}
     * @alias BaseHookClass
     */
    Boot: typeof BaseHookClass;
    options: Required<EggApplicationCoreOptions>;
    readonly messenger: IMessenger;
    agent?: Agent;
    application?: Application;
    loader: EggApplicationLoader;
    /**
     * @class
     * @param {Object} options
     *  - {Object} [type] - type of instance, Agent and Application both extend koa, type can determine what it is.
     *  - {String} [baseDir] - app root dir, default is `process.cwd()`
     *  - {Object} [plugins] - custom plugin config, use it in unittest
     *  - {String} [mode] - process mode, can be cluster / single, default is `cluster`
     */
    constructor(options?: EggApplicationCoreOptions);
    /**
     * @deprecated please use `options` property instead
     */
    get _options(): Required<EggApplicationCoreOptions>;
    protected loadConfig(): Promise<void>;
    protected load(): Promise<void>;
    /**
     * Usage: new ApiClient({ cluster: app.cluster })
     */
    get cluster(): (clientClass: unknown, options?: object) => any;
    /**
     * Wrap the Client with Leader/Follower Pattern
     *
     * @description almost the same as Agent.cluster API, the only different is that this method create Follower.
     *
     * @see https://github.com/node-modules/cluster-client
     * @param {Function} clientClass - client class function
     * @param {Object} [options]
     *   - {Boolean} [autoGenerate] - whether generate delegate rule automatically, default is true
     *   - {Function} [formatKey] - a method to transform the subscription info into a string，default is JSON.stringify
     *   - {Object} [transcode|JSON.stringify/parse]
     *     - {Function} encode - custom serialize method
     *     - {Function} decode - custom deserialize method
     *   - {Boolean} [isBroadcast] - whether broadcast subscription result to all followers or just one, default is true
     *   - {Number} [responseTimeout] - response timeout, default is 3 seconds
     *   - {Number} [maxWaitTime|30000] - leader startup max time, default is 30 seconds
     * @return {ClientWrapper} wrapper
     */
    clusterWrapper(clientClass: unknown, options?: object): any;
    /**
     * print the information when console.log(app)
     * @return {Object} inspected app.
     * @since 1.0.0
     * @example
     * ```js
     * console.log(app);
     * =>
     * {
     *   name: 'mock-app',
     *   env: 'test',
     *   subdomainOffset: 2,
     *   config: '<egg config>',
     *   controller: '<egg controller>',
     *   service: '<egg service>',
     *   middlewares: '<egg middlewares>',
     *   urllib: '<egg urllib>',
     *   loggers: '<egg loggers>'
     * }
     * ```
     */
    inspect(): any;
    toJSON(): any;
    /**
     * http request helper base on {@link httpclient}, it will auto save httpclient log.
     * Keep the same api with `httpclient.request(url, args)`.
     *
     * See https://github.com/node-modules/urllib#api-doc for more details.
     *
     * @param {String} url request url address.
     * @param {Object} options
     * - method {String} - Request method, defaults to GET. Could be GET, POST, DELETE or PUT. Alias 'type'.
     * - data {Object} - Data to be sent. Will be stringify automatically.
     * - dataType {String} - String - Type of response data. Could be `text` or `json`.
     *   If it's `text`, the callback data would be a String.
     *   If it's `json`, the data of callback would be a parsed JSON Object.
     *   Default callback data would be a Buffer.
     * - headers {Object} - Request headers.
     * - timeout {Number} - Request timeout in milliseconds. Defaults to exports.TIMEOUT.
     *   Include remote server connecting timeout and response timeout.
     *   When timeout happen, will return ConnectionTimeout or ResponseTimeout.
     * - auth {String} - `username:password` used in HTTP Basic Authorization.
     * - followRedirect {Boolean} - follow HTTP 3xx responses as redirects. defaults to false.
     * - gzip {Boolean} - let you get the res object when request connected, default false. alias customResponse
     * - nestedQuerystring {Boolean} - urllib default use querystring to stringify form data which don't
     *   support nested object, will use qs instead of querystring to support nested object by set this option to true.
     * - more options see https://github.com/node-modules/urllib
     * @return {Object}
     * - status {Number} - HTTP response status
     * - headers {Object} - HTTP response headers
     * - res {Object} - HTTP response meta
     * - data {Object} - HTTP response body
     *
     * @example
     * ```js
     * const result = await app.curl('http://example.com/foo.json', {
     *   method: 'GET',
     *   dataType: 'json',
     * });
     * console.log(result.status, result.headers, result.data);
     * ```
     */
    curl<T = any>(url: HttpClientRequestURL, options?: HttpClientRequestOptions): Promise<HttpClientResponse<T>>;
    /**
     * Create a new HttpClient instance with custom options
     * @param {Object} [options] HttpClient init options
     */
    createHttpClient(options?: HttpClientOptions): HttpClient;
    /**
     * HttpClient instance
     * @see https://github.com/node-modules/urllib
     * @member {HttpClient}
     */
    get httpClient(): HttpClient;
    /**
     * @deprecated please use httpClient instead
     * @alias httpClient
     * @member {HttpClient}
     */
    get httpclient(): HttpClient;
    /**
     * All loggers contain logger, coreLogger and customLogger
     * @member {Object}
     * @since 1.0.0
     */
    get loggers(): EggLoggers;
    /**
     * Get logger by name, it's equal to app.loggers['name'],
     * but you can extend it with your own logical.
     * @param {String} name - logger name
     * @return {Logger} logger
     */
    getLogger(name: string): EggLogger;
    /**
     * application logger, log file is `$HOME/logs/{appname}/{appname}-web`
     * @member {Logger}
     * @since 1.0.0
     */
    get logger(): EggLogger;
    /**
     * core logger for framework and plugins, log file is `$HOME/logs/{appname}/egg-web`
     * @member {Logger}
     * @since 1.0.0
     */
    get coreLogger(): EggLogger;
    _unhandledRejectionHandler(err: any): void;
    /**
     * dump out the config and meta object
     * @private
     */
    dumpConfigToObject(): {
        config: Record<string, any>;
        meta: Record<string, any>;
    };
    /**
     * save app.config to `run/${type}_config.json`
     * @private
     */
    dumpConfig(): void;
    dumpTiming(): void;
    get [EGG_PATH](): string;
    get config(): EggAppConfig;
    /**
     * app.env delegate app.config.env
     * @deprecated
     */
    get env(): string;
    set env(_: string);
    /**
     * app.proxy delegate app.config.proxy
     * @deprecated
     */
    get proxy(): boolean;
    set proxy(_: boolean);
    /**
     * Create an anonymous context, the context isn't request level, so the request is mocked.
     * then you can use context level API like `ctx.service`
     * @member {String} EggApplication#createAnonymousContext
     * @param {Request} [req] - if you want to mock request like querystring, you can pass an object to this function.
     * @return {Context} context
     */
    createAnonymousContext(req?: any): EggContext;
    /**
     * Run async function in the anonymous context scope
     * @see Context#runInAnonymousContextScope
     * @param {Function} scope - the first args is an anonymous ctx, scope should be async function
     * @param {Request} [req] - if you want to mock request like querystring, you can pass an object to this function.
     */
    runInAnonymousContextScope(scope: (ctx: Context) => Promise<void>, req?: unknown): Promise<void>;
    /**
     * Create egg context
     * @function EggApplication#createContext
     * @param  {Req} req - node native Request object
     * @param  {Res} res - node native Response object
     * @return {Context} context object
     */
    createContext(req: IncomingMessage, res: ServerResponse): Context;
}
declare module '@eggjs/core' {
    interface EggCore {
        inspect(): any;
        get currentContext(): EggContext | undefined;
        ctxStorage: AsyncLocalStorage<EggContext>;
        getLogger(name: string): EggLogger;
        get loggers(): EggLoggers;
        createHttpClient(options?: HttpClientOptions): HttpClient;
        HttpClient: typeof HttpClient;
        get httpClient(): HttpClient;
        curl<T = any>(url: HttpClientRequestURL, options?: HttpClientRequestOptions): Promise<HttpClientResponse<T>>;
        createAnonymousContext(req?: any): EggContext;
        runInAnonymousContextScope(scope: (ctx: Context) => Promise<void>, req?: unknown): Promise<void>;
        readonly messenger: IMessenger;
        Subscription: typeof BaseContextClass;
        BaseHookClass: typeof BaseHookClass;
        Boot: typeof BaseHookClass;
    }
}
